set linesize 32000
set pagesize 1000
set echo on
set termout on
set trimspool on
set serveroutput on size 1000000

spool peexample.log

-- example of partition exchange with following conditions:
-- 
-- Range partitioned table
-- Indexes PK, regular, bitmapped - all local
-- validated FK
-- trigger
-- parallel 8 nologging
--
-- tables:
-- ptab - partitioned table
-- ftab - table related by FK
-- etab - table to be exchanged in
--

-- create table with two partitions

drop table ptab;

CREATE TABLE ptab
(
c1 number,
c2 number, 
c3 number
)
TABLESPACE USERS
NOLOGGING
PARTITION BY RANGE (c2)
(  
  PARTITION p1 VALUES LESS THAN (101),  
  PARTITION p2 VALUES LESS THAN (201)
)
parallel 8;

-- load data both partitions

insert into ptab
select rownum,rownum,1
from dba_objects
where rownum < 201;

commit;

-- create indexes

create unique index ptab_pk
on ptab (c1,c2) tablespace users local nologging noparallel;

create index ptab_ni
on ptab (c2) tablespace users local nologging noparallel;

create bitmap index ptab_bi
on ptab (c3) tablespace users local nologging noparallel;

-- gather stats

execute dbms_stats.gather_table_stats(NULL,'PTAB');

-- show stats

select last_analyzed,num_rows
from user_tables
where table_name='PTAB';

select partition_name,last_analyzed,num_rows
from user_tab_partitions
where table_name='PTAB'
order by partition_name;

select index_name,last_analyzed,num_rows
from user_indexes
where table_name='PTAB'
order by index_name;

select index_name,partition_name,last_analyzed,num_rows,status
from user_ind_partitions
where index_name in
(select index_name
from user_indexes
where table_name='PTAB')
order by index_name,partition_name;

-- verify rowcounts

select count(*) from ptab;
select count(*) from ptab partition(P1);
select count(*) from ptab partition(P2);

-- add PK

ALTER TABLE PTAB ADD (
  CONSTRAINT ptab_pk
  PRIMARY KEY
  (c1,c2)
  USING INDEX LOCAL
  ENABLE VALIDATE);
  
-- add trigger

CREATE OR REPLACE TRIGGER PTAB_TRIG
   BEFORE UPDATE OR INSERT
   ON PTAB
   FOR EACH ROW
BEGIN
   :NEW.C3 := 2;
END;
/
show errors

-- setup FK 

drop table ftab;

CREATE TABLE ftab
(
c3 number
);

insert into ftab values (1);
insert into ftab values (2);

commit;

create unique index ftab_pk
on ftab (c3);

ALTER TABLE FTAB ADD (
  CONSTRAINT ftab_pk
  PRIMARY KEY
  (c3)
  USING INDEX
  ENABLE VALIDATE);

ALTER TABLE ptab ADD (
  CONSTRAINT FK_C3 
  FOREIGN KEY (C3) 
  REFERENCES FTAB (C3)
  ENABLE VALIDATE);
  
-- show constraint status
  
select constraint_name,status,validated
from user_constraints 
where table_name in ('PTAB','FTAB');

-- at this point the environment is setup
-- nothing related to the partition exchange
-- has started yet.  From here the steps
-- to reorg the partitioned table using
-- partition exchange begin.

-- create table to hold one partition's data:

drop table etab;

CREATE TABLE etab
(
c1 number,
c2 number, 
c3 number
)
TABLESPACE USERS
NOLOGGING
parallel 8;

create unique index etab_pk
on etab (c1,c2) tablespace users nologging noparallel;

create index etab_ni
on etab (c2) tablespace users nologging noparallel;

create bitmap index etab_bi
on etab (c3) tablespace users nologging noparallel;

-- load etab with p1 data

-- make  non-pk indexes unusable

alter index etab_ni unusable;
alter index etab_bi unusable;

-- copy data in parallel

alter session enable parallel dml;

insert into etab select * from ptab partition(p1);

commit;

alter session disable parallel dml;

-- rebuild non-pk indexes

alter index etab_ni rebuild parallel 8 nologging;
alter index etab_bi rebuild parallel 8 nologging;

-- back up p1 partition stats

drop table partition_stats;

execute DBMS_STATS.CREATE_STAT_TABLE (NULL,'PARTITION_STATS');

execute DBMS_STATS.EXPORT_TABLE_STATS (NULL,'PTAB','P1',stattab=>'PARTITION_STATS');

-- disable constraints

alter table PTAB modify constraint FK_C3 disable;
alter table PTAB modify constraint PTAB_PK disable;

-- disable trigger

alter trigger PTAB_TRIG disable;

-- exchange etab with ptab partition p1:

alter table ptab 
exchange partition p1 
with table etab
including indexes 
without validation;

-- enable trigger

alter trigger PTAB_TRIG enable;

-- enable constraints novalidate

alter table PTAB modify constraint FK_C3 enable novalidate;
alter table PTAB modify constraint PTAB_PK enable novalidate;

-- restore p1 partition stats

execute DBMS_STATS.IMPORT_TABLE_STATS (NULL,'PTAB','P1',stattab=>'PARTITION_STATS');

-- now the partition exchange is done.
-- dump out stats again to show that everything is the same as it was before

select last_analyzed,num_rows
from user_tables
where table_name='PTAB';

select partition_name,last_analyzed,num_rows
from user_tab_partitions
where table_name='PTAB'
order by partition_name;

select index_name,last_analyzed,num_rows
from user_indexes
where table_name='PTAB'
order by index_name;

select index_name,partition_name,last_analyzed,num_rows,status
from user_ind_partitions
where index_name in
(select index_name
from user_indexes
where table_name='PTAB')
order by index_name,partition_name;

-- verify rowcounts

select count(*) from ptab;
select count(*) from ptab partition(P1);
select count(*) from ptab partition(P2);

-- show constraint status
  
select constraint_name,status,validated
from user_constraints 
where table_name in ('PTAB','FTAB');

spool off
